from .Detector import Detector, TriggeredDetector
import PyTango
import numpy as np

class AdLinkAnalogInput(Detector, TriggeredDetector):
    """
    Interface to the AdLink Tango device exposing one analog input.
    """
    def __init__(self, name=None, device=None):
        self.dev_name = device
        Detector.__init__(self, name=name)
        TriggeredDetector.__init__(self)

    def initialize(self):
        self.dev = PyTango.DeviceProxy(self.dev_name)
        self.dev.init()
        
        self.stop() # needs to be in standby to change anything

        self.dev.write_attribute("TriggerSources", "ExtD:+")
        self.dev.write_attribute("TriggerMode", 1)
        self.dev.write_attribute("NumOfDisplayableTriggers", -1)
        self.dev.write_attribute("ChannelSamplesPerTrigger", 100)

    def prepare(self, acqtime, dataid):
        """
        Run before acquisition, once per scan. Set up triggering,
        number of images etc.
        """

        if self.busy():
            raise Exception('%s is busy!' % self.name)

        if self.hw_trig:
            # fixed at 100 samples per trigger in the Adlink device:
            srate = int(np.ceil(100.0 / acqtime))
            self.dev.write_attribute('NumOfTriggers', self.hw_trig_n)
            self.dev.write_attribute('SampleRate', srate)
        else:
            print('Warning: the AdLinkAnalogInput detector only works with hardware triggering.')

    def arm(self):
        """
        Start the detector if hardware triggered
        """
        if self.busy():
            raise Exception('%s is busy!' % self.name)
        if self.hw_trig:
            self.dev.Start()

    def start(self):
        """
        Start acquisition when software triggered.
        """
        pass

    def stop(self):
        if not self.dev.State() == PyTango.DevState.STANDBY:
            # AdLink gets angry if you stop it in standby
            self.dev.Stop()

    def busy(self):
        return (self.dev.State() == PyTango.DevState.RUNNING)

    def read(self):
        if not self.hw_trig:
            return -1
        vals = self.dev.read_attribute("C00_MeanValues").value
        self.dev.Stop() # have to explicitly call stop after the read
        return np.array(vals)




from .Detector import Detector, SoftwareLiveDetector, Link

import time
import numpy as np
import h5py

class DummyDetector(Detector, SoftwareLiveDetector):
    def __init__(self, name=None):
        Detector.__init__(self, name=name)
        SoftwareLiveDetector.__init__(self)

    def initialize(self):
        pass

    def start(self):
        super(DummyDetector, self).start()
        try:
            self.val = np.random.rand() * self.acqtime
            self._started = time.time()
        except AttributeError:
            raise Exception('Detector not prepared!')

    def stop(self):
        try:
            self._started = time.time() - self.acqtime
        except AttributeError:
            return

    def busy(self):
        try:
            return time.time() < self._started + self.acqtime
        except AttributeError:
            return False

    def read(self):
        try:
            return self.val
        except AttributeError:
            raise Exception('Detector not started!')

class Dummy1dDetector(DummyDetector):
    def start(self):
        super(Dummy1dDetector, self).start()
        try:
            self.val = np.random.rand(100) * self.acqtime
            self._started = time.time()
        except AttributeError:
            raise Exception('Detector not prepared!')

class DummyWritingDetector(DummyDetector):
    def prepare(self, acqtime, dataid):
        super(DummyWritingDetector, self).prepare(acqtime, dataid)
        if dataid is None:
            self.filename_base = None
        else:
            self.filename_base = '/tmp/Dummy_scan_%03d_image_%%03d.hdf5' % dataid
            self.next_image = -1

    def start(self):
        super(DummyWritingDetector, self).start()
        shape = (195, 487)
        if self.filename_base is None:
            self.latest_link = None
        else:
            self.next_image += 1
            filename = self.filename_base % self.next_image
            datapath = 'entry/measurement/data'
            with h5py.File(filename, 'w') as fp:
                fp[datapath] = np.arange(np.prod(shape), dtype=np.int32).reshape(shape)
            self.latest_link = Link(filename, datapath, shape=shape, dtype=np.int32)

    def read(self):
        return self.latest_link

class DummyDictDetector(DummyDetector):
    """
    Illustrates how multi-channel detectors can return dict values.
    """
    def start(self):
        super(DummyDictDetector, self).start()
        try:
            self.val = {'ch1': np.random.rand() * self.acqtime,
                        'ch2': np.random.rand() * self.acqtime * 2,
                        'ch3': np.random.rand() * self.acqtime * 3,}
            self._started = time.time()
        except AttributeError:
            raise Exception('Detector not prepared!')



from .Lima import LimaDetector
from ..environment import env

import numpy as np
import PyTango
import os

class Xspress3(LimaDetector):
    EXT_TRG_MODE = "EXTERNAL_GATE"

    def __init__(self, *args, **kwargs):
        super(Xspress3, self).__init__(*args, **kwargs)
        self._hdf_path_base = 'entry_%04d/measurement/xspress3/data'




import time
import numpy as np
import os, ast

from ..Gadget import Gadget
from ..environment import macro, env
from .. import utils

class MotorLimitException(Exception):
    pass

class Motor(Gadget):
    """
    General base class for motors.

    Motors have dial and user positions, which are related by an offset
    and a scaling factor. The user position is supposed to be used in
    all macros and movements. The dial position is hard coded in the Motor
    subclass to closely reflect what the underlying device quotes.

    user = dial * scaling + offset,
    dial = (user - offset) / scaling

    The user position can be changed, which updates the offset but not
    the scaling. In the same way, limits on the user position are internally
    converted to dial limits, such that setting the user position leaves
    the dial limits unchanged.

    The interface is defined as follows (* means override!):
      user_position (r/w property)  - the current user position.
    * dial_position (r/w property)  - the current dial position, setting
                                      moves the motor with no limit check
      user_limits (r/w property)    - limits in terms of the current
                                      user position settings.
      dial_limits (r/w propery)     - limits on the motor dial position.
      position()                    - returns user_position
      move(pos)                     - move motor to user position pos
                                      while respecting limits
    * busy()                        - motor state
    * stop()                        - halts the motor

    """

    def __init__(self, scaling=1.0, offset=0.0, dial_limits=(None, None),
                 user_format='%.2f', dial_format='%.2f', **kwargs):
        super(Motor, self).__init__(**kwargs)
        self._lowlim, self._uplim = dial_limits
        self._scaling = scaling
        self._offset = offset
        self._uformat = user_format
        self._dformat = dial_format

    @property
    def user_position(self):
        return self.dial_position * self._scaling + self._offset

    @user_position.setter
    def user_position(self, pos):
        self._offset = pos - self.dial_position * self._scaling

    @property
    def user_limits(self):
        if None not in (self._uplim, self._lowlim):
            l1 = self._lowlim * self._scaling + self._offset
            l2 = self._uplim * self._scaling + self._offset
        else:
            low, up = None, None
        return tuple(sorted([l1, l2]))

    @user_limits.setter
    def user_limits(self, lims):
        l1 = (lims[0] - self._offset) / self._scaling
        l2 = (lims[1] - self._offset) / self._scaling
        self._lowlim, self._uplim = sorted([l1, l2])

    @property
    def dial_limits(self):
        return (self._lowlim, self._uplim)

    @dial_limits.setter
    def dial_limits(self, lims):
        self._lowlim, self._uplim = lims

    def position(self):
        return self.user_position

    def move(self, pos):
        if self.busy():
            raise Exception('Motor is busy')
        dial = (pos - self._offset) / self._scaling
        try:
            assert dial <= self._uplim
            assert dial >= self._lowlim
        except AssertionError:
            print('Trying to move %s outside its limits!' % self.name)
            return -1
        except TypeError:
            pass
        self.dial_position = dial

    @property
    def dial_position(self):
        """
        Override this! Reads the dial position.
        """
        raise NotImplementedError

    @dial_position.setter
    def dial_position(self, pos):
        """
        Override this! Move the motor to dial position.
        """
        raise NotImplementedError

    def busy(self):
        """
        Overrides this! Determines if the motor is busy.
        """
        raise NotImplementedError

    def stop(self):
        """
        Override this! Stops the motor.
        """
        raise NotImplementedError


class DummyMotor(Motor):
    def __init__(self, *args, **kwargs):
        super(DummyMotor, self).__init__(*args, **kwargs)
        self._aim = 0.0
        self._oldpos = 0.0
        self._started = 0.0
        self.velocity = 1.0

    @property
    def dial_position(self):
        dpos = self._aim - self._oldpos
        dt = time.time() - self._started
        T = abs(dpos / self.velocity)
        if dt < T:
            return self._oldpos + dpos * self.velocity * dt / T
        else:
            return self._aim

    @dial_position.setter
    def dial_position(self, pos):
        self._oldpos = self.position()
        self._started = time.time()
        self._aim = pos

    def busy(self):
        return not np.isclose(self._aim, self.dial_position)

    def stop(self):
        self._aim = self.dial_position


class MotorMemorizer(Gadget):
    """
    Memorizes limits and offset values for Motors.
    """
    def __init__(self, filepath=None, **kwargs):
        super(MotorMemorizer, self).__init__(**kwargs)
        self.filepath = filepath
        if filepath and os.path.exists(filepath):
            self.load()

    def load(self):
        try:
            motors = {m.name: m for m in Motor.getinstances()}
            with open(self.filepath, 'r') as fp:
                for row in fp:
                    dct = ast.literal_eval(row)
                    if dct['name'] in motors:
                        motors[dct['name']]._offset = dct['_offset']
                        if dct['_lowlim']:
                            motors[dct['name']]._lowlim = dct['_lowlim']
                        if dct['_uplim']:
                            motors[dct['name']]._uplim = dct['_uplim']
            print('Loaded motor states from %s' % self.filepath)
        except (FileNotFoundError,):
            print("Memorizer file %s doesn't exist" % self.filepath)

    def dump(self):
        try:
            with open(self.filepath, 'w') as fp:
                for m in Motor.getinstances():
                    dct = {'name': m.name,
                           '_offset': m._offset,
                           '_lowlim': m._lowlim, '_uplim': m._uplim}
                    fp.write(str(dct) + '\n')
            print('Saved motor states to %s' % self.filepath)
        except (FileNotFoundError,):
            print("Cant write to %s, doesn't exist")


@macro
class Mv(object):
    """
    Move one or more motors.

    mvr <motor1> <position1> <motor2> <position2> ...

    """
    def __init__(self, *args):
        self.motors = args[::2]
        self.targets = np.array(args[1::2])

    def _run_while_waiting(self):
        pass
    
    def run(self):
        if max(m.userlevel for m in self.motors) > env.userLevel:
            print('You are trying to move motors above your user level!')
            return
        for m, pos in zip(self.motors, self.targets):
            m.move(pos)
        try:
            while True in [m.busy() for m in self.motors]:
                self._run_while_waiting()
                time.sleep(.01)
        except KeyboardInterrupt:
            for m in self.motors:
                m.stop()

@macro
class Mvd(object):
    """
    Move one or more motors to an abolute dial position.
    """
    def run(self):
        raise NotImplementedError

@macro
class Mvr(Mv):
    """
    Move one or more motors relative to their current positions.

    mvr <motor1> <position1> <motor2> <position2> ...

    """
    def __init__(self, *args):
        self.motors = args[::2]
        displacements = np.array(args[1::2])
        current = np.array([m.position() for m in self.motors])
        self.targets = current + displacements

@macro
class Umv(Mv):
    """
    Like mv, except it prints the positions while moving.
    """
    def _run_while_waiting(self):
        l = ['%s: %.2f' % (m.name, m.user_position) for m in self.motors]
        s = '; '.join(l)
        print(s + '\r', end='')

    def run(self):
        # ensures the final position is printed too
        super(Umv, self).run()
        self._run_while_waiting()

@macro
class Umvr(Mvr, Umv):
    """
    Like mvr, except it prints the positions while moving.
    """
    pass # less is more

@macro
class Wm(object):
    """
    Print the positions of one or more motors.

    wm <motor1> <motor2> ...
    """
    def __init__(self, *args):
        self.motors = args
    def run(self, *args):
        titles = ['motor', 'user pos.', 'limits', 'dial pos.', 'limits']
        table = []
        for m in self.motors:
            try:
                upos = m._uformat % m.user_position
                dpos = m._dformat % m.dial_position
                if None in (m._uplim, m._lowlim):
                    ulims = '(None, None)'
                    dlims = '(None, None)'
                else:
                    ulims = ('(%s, %s)' % (2*(m._uformat,))) % m.user_limits
                    dlims = ('(%s, %s)' % (2*(m._dformat,))) % m.dial_limits
                table.append([m.name, upos, ulims, dpos, dlims])
            except:
                print('Could not get position of %s' % m.name)
        print(utils.list_to_table(lst=table, titles=titles, margins=[5,2,5,2,0]))

@macro
class Wa(Wm):
    """
    Print the positions of all motors available at the current user level.
    """
    def __init__(self):
        self.motors = [m for m in Motor.getinstances()
                       if m.userlevel <= env.userLevel]

@macro
class LsM(object):
    """
    List available motors.
    """
    def run(self):
        dct = {m.name: m.__class__ for m in Motor.getinstances()
               if m.userlevel <= env.userLevel}
        print(utils.dict_to_table(dct, titles=('name', 'class'), sort=True))

@macro
class SetLim(object):
    """
    Set limits on motors.

    setlim <motor1> <lower 1> <upper 1> ...
    """
    def __init__(self, *args):
        self.motors = args[::3]
        self.lowers = args[1::3]
        self.uppers = args[2::3]

    def run(self):
        for m, l, u in zip(self.motors, self.lowers, self.uppers):
            m.user_limits = (l, u)

        # memorize the new state
        for m in MotorMemorizer.getinstances():
            m.dump()

@macro
class SetPos(object):
    """
    Sets user position on motors.

    setpos <motor1> <pos1> ...
    """
    def __init__(self, *args):
        self.motors = args[::2]
        self.positions = args[1::2]

    def run(self):
        for m, p in zip(self.motors, self.positions):
            m.user_position = p

        # memorize the new state
        for m in MotorMemorizer.getinstances():
            m.dump()



import time
import numpy as np
from ..environment import macro, env
from ..recorders import active_recorders, RecorderHeader, RecorderFooter
from ..detectors import Detector

class SoftwareScan(object):
    """
    Base class for the normal sardana-style software-controlled scan. Respects
    the availability and deadlines managed by env.scheduler.

    Overwrite these methods:
        __init__ (which reads in the scan parameters) and
        _generate_positions (which generates the line scan, mesh, spiral, or what ever you like)
    """

    def __init__(self, exposuretime):
        """
        The constructor should parse parameters.
        """
        self.motors = []   # list of motors, to be filled by subclass
        self.exposuretime = exposuretime
        self.scannr = env.nextScanID
        env.nextScanID += 1

    def header_line(self):
        motor_names = [m.name for m in self.motors]
        group = Detector.get_active_detectors()
        det_names = [d.name for d in group]
        header = '#       ' + len(motor_names) * '%-8s' + len(group) * '%-14s' + 'dt'
        line = '-' * (8 + len(motor_names) * 8 + len(group) * 16 + 8)
        return ('\n' + header + '\n' + line) % tuple(motor_names + det_names)

    def table_line(self):
        group = Detector.get_active_detectors()
        return '%-8d' + len(self.motors) * '%-8.2f' + len(group) * '%-12s  ' + '%-8.2f'

    def format_number(self, nr):
        if isinstance(nr, dict):
            return '<%d-dict>' % len(nr)
        try:
            return '%8s' % (nr.shape,)
        except:
            pass
        try:
            return '%8e' % nr
        except TypeError:
            return str(nr)[-8:]

    def _calc_time_needed(self):
        """
        Suggestion for how to organize topup avoidance. You could 
        also look at how long acquisitions have tended to take. Or
        you could try to estimate the time needed based on the
        input parameters.
        """
        return self.exposuretime * 5 + 5

    def _before_move(self):
        """
        Gets called for each step.
        """

        # Example implementation of topup avoidance
        time_needed = self._calc_time_needed()
        enough_time = time_needed < env.scheduler.limit if env.scheduler.limit else True
        ready = env.scheduler.ready
        while not ready or not enough_time:
            if not enough_time:
                print('not enough time to complete this measurement, so waiting...')
            else:
                print('Waiting for beamline to become available...')
            time.sleep(1.)
            enough_time = time_needed < env.scheduler.limit
            ready = env.scheduler.ready

    def _before_arm(self):
        """
        Gets called for each step.
        """
        pass

    def _while_acquiring(self):
        """
        Gets called repeatedly while the detectors detect.
        """
        pass

    def _before_start(self):
        """
        Gets called for each step.
        """
        pass

    def run(self):
        """
        This method does all the serious interaction with motors,
        detectors, and data recorders.
        """
        print('\nScan #%d starting at %s' % (self.scannr, time.asctime()))
        print(self.header_line())
        positions = self._generate_positions()
        table_line = self.table_line()
        # find and prepare the detectors
        group = Detector.get_active_detectors()
        if group.busy():
            print('Detectors are busy: %s' % (', '.join([d.name for d in group if d.busy()])))
            return
        group.prepare(self.exposuretime, self.scannr)
        t0 = time.time()
        # send a header to the recorders
        snap = env.snapshot.capture()
        for r in active_recorders():
            r.queue.put(RecorderHeader(scannr=self.scannr, path=env.paths.directory, snapshot=snap))
        try:
            for i, pos in enumerate(positions):
                # move motors
                self._before_move()
                for m in self.motors:
                    m.move(pos[m.name])
                while True in [m.busy() for m in self.motors]:
                    time.sleep(.01)
                # arm detectors
                self._before_arm()
                group.arm()
                # start detectors
                self._before_start()
                group.start()
                while group.busy():
                    self._while_acquiring()
                    time.sleep(.01)
                # read detectors and motors
                dt = time.time() - t0
                dct = {'dt': dt}
                for d in group:
                    dct[d.name] = d.read()
                for m in self.motors:
                    dct[m.name] = m.position()
                # pass data to recorders
                for r in active_recorders():
                    r.queue.put(dct)
                # print spec-style info
                print(table_line % tuple([i] + 
                      [m.position() for m in self.motors]
                      + [self.format_number(dct[d.name]) for d in group] 
                      + [dt]))
            print('\nScan #%d ending at %s' % (self.scannr, time.asctime()))
        except KeyboardInterrupt:
            group.stop()
            print('\nScan #%d cancelled at %s' % (self.scannr, time.asctime()))

        # tell the recorders that the scan is over
        for r in active_recorders():
            r.queue.put(RecorderFooter())
        
    def _generate_positions(self):
        """
        Function or generator which returns or yields an iterable of
        dicts {motorA.name: posA, motorB.name: posB, ...}
        """
        raise NotImplementedError

@macro
class LoopScan(SoftwareScan):
    """
    A software scan with no motor movements. Number of exposures is
    <intervals> + 1, for consistency with ascan, dscan etc.

    loopscan <intervals> <exp_time>
    """
    def __init__(self, intervals, exposuretime=1.0):
        """
        Parse arguments.
        """
        super(LoopScan, self).__init__(float(exposuretime))
        self.intervals = intervals
        self.motors = []

    def _generate_positions(self):
        # dummy positions with a non existent motor
        for i in range(self.intervals + 1):
            yield {'fake':i}

@macro
class Ct(object):
    """
    Make a single acquisition on the active detectors without recording
    data. Optional argument <exp_time> specifies exposure time, the default
    is 1.0.

    ct [<exp_time>]
    """
    def __init__(self, exp_time=1, print_nd=True):
        """
        Parse arguments.
        """
        self.exposuretime = float(exp_time)

    def run(self):
        # find and prepare the detectors
        group = Detector.get_active_detectors()
        group.prepare(self.exposuretime, dataid=None)
        # arm and start detectors
        group.arm()
        while group.busy():
            time.sleep(.01)
        group.start()
        while group.busy():
            time.sleep(.01)
        # read detectors and motors
        dct = {}
        for d in group:
            dct[d.name] = d.read()
        # print results
        for key, val in dct.items():
            try:
                val_ = val.shape
            except AttributeError:
                val_ = val
            print(key, ':', val_)



"""
Provides a general Motor interface to any Tango attribute.
"""

import PyTango
from . import Motor

class TangoAttributeMotor(Motor):
    """
    Motor interface to any Tango attribute, so that anything can
    be scanned. These motors cannot be stopped and are never
    considered busy.
    """

    def __init__(self, device, attribute, **kwargs):
        super(TangoAttributeMotor, self).__init__(**kwargs)
        self.proxy = PyTango.DeviceProxy(device)
        self.attribute = attribute

    @property
    def dial_position(self):
        return self.proxy.read_attribute(self.attribute).value

    @dial_position.setter
    def dial_position(self, pos):
        self.proxy.write_attribute(self.attribute, pos)

    def busy(self):
        return False

    def stop(self):
        pass



from ..environment import macro, MacroSyntaxError
from .Mesh import Mesh
import tty, sys, termios
from ..motors import all_are_motors

# constants to keep track of key buttons
KEY_UP = '\x1b[A'
KEY_DOWN = '\x1b[B'
KEY_RIGHT = '\x1b[C'
KEY_LEFT = '\x1b[D'

def getarrowkey():
    """
    A function which waits for a keypress and returns one of the above constants.
    """
    ch = None
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setcbreak(sys.stdin.fileno())
        ch = sys.stdin.read(3)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
    if ch in (KEY_UP, KEY_DOWN, KEY_RIGHT, KEY_LEFT):
        return ch
    else:
        return None

@macro
class Tweak(Mesh):
    """
    An interactive scan where motor positions are chosen manually for
    each point. Useful for tweaking motors and reading the currently
    active detectors after each step.

    tweak <motor1> <stepsize1> [<motor2> <stepsize2>] <exp_time>
    """

    def __init__(self, *args):
        """
        Parse arguments.
        """
        try:
            exposuretime = float(args[-1])
            super(Mesh, self).__init__(exposuretime)
            self.motors = args[:-1:2]
            self.steps = args[1::2]
            assert len(args) in (3, 5)
            assert all_are_motors(self.motors)
        except:
            raise
            raise MacroSyntaxError
        print('\nUse the arrow keys to tweak motors and ctrl-C to stop.')

    def _generate_positions(self):
        positions = {m.name:m.position() for m in self.motors}
        yield positions
        while True:
            key = getarrowkey()
            imotor = 0 if key in (KEY_LEFT, KEY_RIGHT) else -1
            direction = -1 if key in (KEY_DOWN, KEY_LEFT) else 1
            positions[self.motors[imotor].name] += direction * self.steps[imotor]
            yield positions

    def _before_move(self):
        # Override so as not to respect the scheduler.
        pass



from .Detector import Detector, DetectorGroup, LiveDetector, Link, TriggeredDetector
from .Dummies import DummyDetector, Dummy1dDetector, DummyWritingDetector, DummyDictDetector

# don't automatically import hardware-specific classes here,
# as these might have special dependencies that aren't
# available everywhere.



"""
Module which provides Scheduler classes, which help the acquisition
system determine whether the instrument is ready to acquire, and if
there are any deadlines coming up that have to be avoided.
"""

import time

class DummyScheduler(object):
    """
    Dummy base class to define the API.
    """
    @property
    def ready(self):
        """
        Whether or not the system is available now.
        """
        return self._ready()

    @property
    def limit(self):
        """
        Time left (in seconds) until the next deadline, None for no deadline.
        """
        return self._limit()

    def _ready(self):
        return True

    def _limit(self):
        return None


class DummyInjectionScheduler(DummyScheduler):
    """
    Dummy that does injection for one minute every five minutes.
    """
    def _ready(self):
        return time.localtime().tm_min % 5

    def _limit(self):
        whole_mins = 4 - (time.localtime().tm_min % 5)
        whole_secs = 60 - time.localtime().tm_sec
        return whole_mins * 60 + whole_secs


class MaxivInjectionScheduler(DummyScheduler):
    """
    Scheduler which keeps track of storage ring injections at MAX IV.
    Could also check the shutter status, so that closing the safety
    shutter pauses the scan!

    NOTE: It's probably best to write this as an asynchronous
    thread or process, because it might involve reading network
    attributes from the other side of the lab. It could do this at
    1 Hz or so, and the getters could just report on the latest
    value. That way scanning won't be slowed down.
    """
    pass



from . import Recorder
import h5py
import time
import numpy as np
import os

class Hdf5Recorder(Recorder):
    """
    Recorder which writes to hdf5 files.
    """
    def __init__(self, name=None):
        Recorder.__init__(self, name=name)
        self.fp = None

    def act_on_header(self, dct):
        """
        Opens a file when a new scan starts.
        """
        filename = os.path.join(dct['path'], '%06u.h5'%dct['scannr'])
        if os.path.isfile(filename):
            print('************ WARNING ************')
            print('Data already exists! Hdf5Recorder')
            print('won''t write data to this target.')
            print('*********************************')
            self.fp = None
        else:
            self.fp = h5py.File(filename, 'w')
            self.act_on_data({'snapshot':dct['snapshot']}, base='entry/')

    def act_on_data(self, dct, base='entry/measurement/'):
        """
        Write data packets to the h5 file.
        """
        if self.fp is None:
            print('** no hdf5 file open, so not writing anything')
            return
        for key, val in dct.items():
            name = base + key

            # treat dict values recursively
            if type(val) == dict:
                new_dct = {key + '/' + str(k): v for k, v in val.items()}
                self.act_on_data(new_dct, base=base)
                continue

            # decide where to put the new data
            if not name in self.fp:
                # create datasets the first time around
                if isinstance(val, np.ndarray):
                    # arrays
                    d = self.fp.create_dataset(name, shape=(1,)+val.shape, maxshape=(None,)+val.shape, dtype=val.dtype)
                elif isinstance(val, h5py.ExternalLink):
                    # links
                    layout = h5py.VirtualLayout(shape=(1,)+val.shape, maxshape=(None,)+val.shape, dtype=val.dtype)
                    self.fp.create_virtual_dataset(name, layout)
                elif (type(val) == str):
                    # strings
                    d = self.fp.create_dataset(name, shape=(1,), maxshape=(None,), dtype='S100')
                else:
                    # scalars of any type
                    d = self.fp.create_dataset(name, shape=(1,), maxshape=(None,), dtype=type(val))
            else:
                # then resize datasets
                d = self.fp[name]
                d.resize((d.shape[0]+1,) + d.shape[1:])

            # special case: convert strings
            if type(val) == str:
                val = val.encode(encoding='ascii', errors='ignore')

            # write the data
            if isinstance(val, h5py.ExternalLink):
                vsource = h5py.VirtualSource(val.filename, val.path, shape=val.shape)
                d[-1] = vsource
            else:
                # a dataset
                d[-1] = val

    def act_on_footer(self):
        """
        Closes the file after the scan.
        """
        if self.fp is not None:
            self.fp.flush()
            self.fp.close()



from .Gadget import Gadget
from collections import OrderedDict
from fnmatch import filter

def list_to_table(lst, titles, margins=3, sort=True):
    """
    Formats a table from a nested list, where the first index is the row.
    """
    if sort:
        lst = sorted(lst)
    result = ''
    margins = [margins,] * len(titles) if not hasattr(margins, '__iter__') else margins
    # establish column widths
    widths = []
    for i in range(len(titles)):
        widths.append(max([len(titles[i]),] + [len(row[i]) for row in lst]) + margins[i])
    # a base format string for every line
    linebase = ''
    for w in widths:
        linebase += ('%%-%ss'%w)
    # make the header
    result += linebase % tuple(titles) + '\n'
    result += '-' * sum(widths) + '\n'
    # add the table data
    for row in lst:
        result += linebase % tuple(row) + '\n'
    return result.strip()

def dict_to_table(dct, titles=('col1', 'col2'), margin=3, sort=True):
    """
    Formats a dict where key:val is str:str into a two column table.
    """
    if sort:
        dct = OrderedDict({key:dct[key] for key in sorted(dct.keys())})
    col1width = max([len(str(s)) for s in dct.keys()] + [len(titles[0])])
    col2width = max([len(str(s)) for s in dct.values()] + [len(titles[1])])
    width = col1width + col2width + margin
    baseline = ('%%-%ds' % col1width) + ' ' * margin + ('%%-%ds' % col2width)
    output = '\n' + baseline % titles + '\n' + '-' * width
    for key, val in dct.items():
        output += '\n' + (baseline % (key, val))
    return output

def str_to_args(line):
    """
    Handy function which splits a list of arguments, translates
    names of Gadget instances to actual objects, and evaluates
    other expressions. For example,

    In [11]: str_to_args("samx 'hej' 1./20")
    Out[11]: [<lib.motors.Motor.DummyMotor at 0x7f6b11b0ba90>, 'hej', 0.05]

    """
    args_in = line.split()
    args_out = []
    gadget_lookup = {g.name: g for g in Gadget.getinstances()}
    for a in args_in:
        if ('*' in a) or ('?' in a):
            matching_names = filter(gadget_lookup.keys(), a)
            args_out += [gadget_lookup[name] for name in matching_names]
        elif a in gadget_lookup.keys():
            args_out.append(gadget_lookup[a])
        else:
            args_out.append(eval(a))
    return args_out




from .Scan import SoftwareScan, Ct, LoopScan
from .Mesh import Mesh, DMesh
from .AScan import AScan, DScan
from .Tweak import Tweak
from .NpointFlyscan import NpointFlyscan
from .Spiral import SpiralScan



"""
Module which contains classes that gather information which
needs to be dumped to every scan.
"""

class EmptySnapshot(object):
    """
    Define the API.
    """
    def capture(self):
        """
        Returns a dict of label-value pairs.
        """
        return {}

class MotorSnapshot(EmptySnapshot):
    """
    Snapshot which consists of all motor values.
    """
    def capture(self):
        from lib.motors import Motor
        dct = {}
        for m in Motor.getinstances():
            try:
                dct[m.name] = m.position()
            except:
                print('Could not take snapshot of motor %s' % m.name)
        return dct



from . import Motor

class PseudoMotor(Motor):
    """
    Pseudo motor base class. Subclass and override the following:
           
          calc_pseudo(self, physicals)
          calc_physicals(self, pseudo, current_physicals)
    """
    def __init__(self, motors, dry_run=False, *args, **kwargs):
        super(PseudoMotor, self).__init__(*args, **kwargs)
        self.motors = motors
        self.dry_run = dry_run

    def physicals(self):
        return [m.position() for m in self.motors]

    @property
    def dial_position(self):
        return self.calc_pseudo(self.physicals())

    @dial_position.setter
    def dial_position(self, pos):
        physicals = self.calc_physicals(pos, self.physicals())
        for m, pos in zip(self.motors, physicals):
            if self.dry_run:
                print('Would move %s to %f' % (m.name, pos))
            else:
                m.move(pos)

    def busy(self):
        return True in [m.busy() for m in self.motors]

    def stop(self):
        [m.stop for m in self.motors()]

    def calc_pseudo(self, physicals):
        raise NotImplementedError

    def calc_physicals(self, pseudo, current_physicals):
        raise NotImplementedError

class ExamplePseudoMotor(PseudoMotor):
    """
    Difference between two motors
    """
    def calc_pseudo(self, physicals):
        return physicals[1] - physicals[0]

    def calc_physicals(self, pseudo, current_physicals):
        current_diff = self.calc_pseudo(current_physicals)
        half_increase = (pseudo - current_diff) / 2.0
        return [current_physicals[0] - half_increase, 
                current_physicals[1] + half_increase]




from .Scan import SoftwareScan
from ..environment import macro, MacroSyntaxError
from ..motors import all_are_motors
import numpy as np
import time

@macro
class Mesh(SoftwareScan):
    """
    Software scan on a regular grid of N motors.
        
    mesh <motor1> <start> <stop> <intervals> ... <exp_time>
    """

    def __init__(self, *args):
        """
        Parse arguments.
        """
        self.motors = []
        self.limits = []
        self.intervals = []
        try:
            exposuretime = float(args[-1])
            super(Mesh, self).__init__(exposuretime)
            for i in range(int((len(args) - 1) / 4)):
                self.motors.append(args[4*i])
                self.limits.append([float(m) for m in args[4*i+1:4*i+3]])
                self.intervals.append(int(args[4*i+3]))
            assert all_are_motors(self.motors)
            assert (len(args) - 1) % 4 == 0
        except:
            raise MacroSyntaxError

    def _generate_positions(self):
        positions = []
        for i in range(len(self.motors)):
            positions.append(np.linspace(self.limits[i][0],
                                         self.limits[i][1],
                                         self.intervals[i]+1))
        grids = np.meshgrid(*reversed(positions))
        grids = [l for l in reversed(grids)] # fastest last
        for i in range(len(grids[0].flat)):
            yield {m.name: pos.flat[i] for (m, pos) in zip(self.motors, grids)}

@macro
class DMesh(Mesh):
    """
    Software scan on a regular grid of N motors, with positions relative
    to each motor's current one. Moves motors back at the end.

    dmesh <motor1> <start> <stop> <intervals> ... <exp_time>
    """
    def _generate_positions(self):
        current = {m.name:m.position() for m in self.motors}
        for pos in super(DMesh, self)._generate_positions():
            for i, m in enumerate(self.motors):
                pos[m.name] += current[m.name]
            yield pos

    def run(self):
        old_pos = [m.position() for m in self.motors]
        super(DMesh, self).run()
        # wait for motors then move them back
        while True in [m.busy() for m in self.motors]:
            time.sleep(.01)
        print('Returning motors to their starting positions...')
        for m, pos in zip(self.motors, old_pos):
            m.move(pos)
        while True in [m.busy() for m in self.motors]:
            time.sleep(.01)
        print('...done')



from .Scan import SoftwareScan
from .AScan import DScan
from ..environment import macro, MacroSyntaxError
from ..motors import all_are_motors
import numpy as np
import time

@macro
class SpiralScan(DScan):
    """
    Software scan across a 2D Archimedes spiral centered on the 
    current position.
        
    spiralscan <motor1> <motor2> <stepsize> <positions> <exp_time>
    """

    def __init__(self, m1, m2, stepsize, npos, exptime):
        """
        Parse arguments. We're inheriting DScan to get its nice run()
        method, but we'll call the SoftwareScan constructor anyway bacause
        we're not interested in DScan's way of parsing arguments.
        """
        try:
            SoftwareScan.__init__(self, float(exptime))
            self.motors = [m1, m2]
            self.stepsize = float(stepsize)
            self.npos = int(npos)
            assert all_are_motors(self.motors)
        except:
            raise MacroSyntaxError

    def _generate_positions(self):
        starting = [m.position() for m in self.motors]
        for t in range(self.npos):
            A = self.stepsize * np.sqrt(t/np.pi)
            B = np.sqrt(4*np.pi*t)
            yield {self.motors[0].name: starting[0] + A * np.cos(B),
                   self.motors[1].name: starting[1] + A * np.sin(B)}



from .Lima import LimaDetector
from ..environment import env

import numpy as np
import PyTango
import os

class Pilatus(LimaDetector):
    def __init__(self, *args, **kwargs):
        super(Pilatus, self).__init__(*args, **kwargs)
        self._hdf_path_base = 'entry_%04d/measurement/Pilatus/data'

    def _initialize_det(self):
        self.energy = 10.0

    @property
    def energy(self):
        return self.det.energy_threshold
        
    @energy.setter
    def energy(self, value):
        if value < 4.5 or value > 36:
            raise ValueError('Requested value is outside the Pilatus range of 4.5-36 keV')
        self.det.write_attribute('energy_threshold', value)




from ..Gadget import Gadget
from .. import utils
from ..environment import macro

from multiprocessing import get_context
# Fancy multiprocessing contexts needed or we will crash matplotlib
# and ipython somehow. On the other hand, 'spawn' is more sensitive
# than 'fork' so you have to be careful only to call Recorder.start()
# from main, see the example scripts.
ctx = get_context('spawn')
Process = ctx.Process
Queue = ctx.Queue

import time
import signal

class RecorderHeader(dict):
    """
    Helper class to define a specific dict format to send recorders
    when a new scan starts.
    """
    def __init__(self, scannr, path, snapshot=None):
        super(RecorderHeader, self).__init__(scannr=scannr, path=path, snapshot=snapshot)

class RecorderFooter(dict):
    """
    Helper class to define a specific dict format to send recorders
    when a new scan finishes.
    """
    pass

class Recorder(Gadget, Process):
    """
    Base class for Recorders. Provides the multiprocessing and queuing
    functionality.
    """
    def __init__(self, name=None, delay=.1):
        Process.__init__(self)
        Gadget.__init__(self, name=name)
        ctx = get_context('spawn')
        self.queue = ctx.Queue()
        self.delay = delay
        self.quit = False

    def _process_queue(self):
        dcts = [self.queue.get() for i in range(self.queue.qsize())] # ok since only we are reading from self.queue
        for dct in dcts:
            if dct is None:
                self.quit = True
            elif isinstance(dct, RecorderHeader):
                self.act_on_header(dct)
            elif isinstance(dct, RecorderFooter):
                self.act_on_footer()
            else:
                self.act_on_data(dct)

    def run(self):
        # ignore SIGINT signals from ctrl-C in the main process
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        self.init()
        while not self.quit:
            time.sleep(self.delay)
            self._process_queue()
            self.periodic_check()
        self._close()

    def init(self):
        """
        Subclass this. Initialize (open windows, open files, ...)
        """
        pass

    def act_on_header(self, dct):
        """
        Subclass this. Performs an action when a new scan is started.
        The key-value pairs of dct are "path":path and "scannr":scannr.
        """
        pass

    def act_on_data(self, dct):
        """
        Subclass this. Performs an action when a new data package is
        received. The keys of dct are detector and motor names, the values
        are their readings at one particular point in the scan.
        """
        pass

    def act_on_footer(self):
        """
        Subclass this. Performs an action when a scan ends.
        """
        pass

    def periodic_check(self):
        """
        Subclass if you like. A function which gets called on every iteration
        of the recorder. Useful for example for checking if files should be
        closed or whether a plot window still exists.
        """

    def _close(self):
        """
        Subclass this. Clean up (close windows, close files...)
        """
        pass

    def stop(self):
        """
        Stop a started subprocess safely by putting a poison pill in
        its queue.
        """
        self.queue.put(None)


class DummyRecorder(Recorder):
    def act_on_header(self, dct):
        print('got a header! am told to write scan %d to %s' % (dct['scannr'], dct['path']))
    def act_on_data(self, dct):
        print('found this!', dct)
    def act_on_footer(self):
        print('scan over, apparently')
    def init(self):
        print('opening')
    def _close(self):
        print('closing')


def active_recorders():
    return [r for r in Recorder.getinstances() if r.is_alive()]

@macro
class LsRec(object):
    """
    List active recorders.
    """
    def run(self):
        dct = {r.name: r.__class__ for r in active_recorders()}
        print(utils.dict_to_table(dct, titles=('name', 'class')))



from .Recorder import Recorder, DummyRecorder, active_recorders, RecorderHeader, RecorderFooter
try:
    from .PlotRecorder import PlotRecorder
except ImportError:
    print('Note: PlotRecorder not available, probably because matplotlib is missing.')

try:
    from .Hdf5Recorder import Hdf5Recorder
except ImportError:
    print("Note: Hdf5Recorder not available, probably because h5py is missing.")




from .Scan import SoftwareScan
from ..environment import macro, MacroSyntaxError
from ..motors import all_are_motors
import numpy as np
import time

@macro
class AScan(SoftwareScan):
    """
    Software scan one or more motors in parallel.
        
    ascan <motor1> <start> <stop> ... <intervals> <exp_time>
    """

    def __init__(self, *args):
        """
        Parse arguments.
        """
        self.motors = []
        self.limits = []
        try:
            exposuretime = float(args[-1])
            self.intervals = int(args[-2])
            super(AScan, self).__init__(exposuretime)
            for i in range(int((len(args) - 2) / 3)):
                self.motors.append(args[3*i])
                self.limits.append([float(m) for m in args[3*i+1:3*i+3]])
            assert all_are_motors(self.motors)
            assert (len(args) - 2) % 3 == 0
        except:
            raise MacroSyntaxError

    def _generate_positions(self):
        positions = []
        for i in range(len(self.motors)):
            positions.append(np.linspace(self.limits[i][0],
                                         self.limits[i][1],
                                         self.intervals+1))
        for i in range(len(positions[0])):
            yield {m.name: pos[i] for (m, pos) in zip(self.motors, positions)}

@macro
class DScan(AScan):
    """
    Software scan one or more motors in parallel, with positions
    relative to each motor's current one. Moves back afterwards.

    dscan <motor1> <start> <stop> <intervals> ... <exp_time>
    """
    def _generate_positions(self):
        current = {m.name:m.position() for m in self.motors}
        for pos in super(DScan, self)._generate_positions():
            for i, m in enumerate(self.motors):
                pos[m.name] += current[m.name]
            yield pos

    def run(self):
        old_pos = [m.position() for m in self.motors]
        super(DScan, self).run()
        # wait for motors then move them back
        while True in [m.busy() for m in self.motors]:
            time.sleep(.01)
        print('Returning motors to their starting positions...')
        for m, pos in zip(self.motors, old_pos):
            m.move(pos)
        while True in [m.busy() for m in self.motors]:
            time.sleep(.01)
        print('...done')



from .Mesh import Mesh
from ..motors import all_are_motors
from ..environment import macro, MacroSyntaxError
from ..detectors import Detector, TriggeredDetector

@macro
class NpointFlyscan(Mesh):
    """
    Flyscan macro for the Npoint LC400 setup.
        
    npointflyscan <fly-motor> <start> <stop> <intervals> <step-motor1> <start1> <stop1> <intervals1> ... <exp time> <latency>

    The argument fly-motor must be an instance of the LC400Motor class.
    """

    def __init__(self, *args):
        """
        Parse arguments.
        """
        try:
            assert all_are_motors(args[:-2:4])
            super(NpointFlyscan, self).__init__(*args[4:-1])
            self.fastmotor = args[0]
            self.fastlimits = [float(i) for i in args[1:3]]
            self.fastintervals = int(args[3])
            self.exptime = float(args[-2])
            self.latency = float(args[-1])
        except:
            #raise MacroSyntaxError
            raise

    def _set_det_trig(self, on):
        for d in Detector.get_active_detectors():
            if isinstance(d, TriggeredDetector):
                d.hw_trig = on
                d.hw_trig_n = self.fastintervals + 1

    def _set_det_active(self, dets, on):
        for d in dets:
            d.active = on
            verb = {True: 'Activated', False: 'Deactivated'}[on]
            print('%s the position buffer detector %s' % (verb, d.name))

    def run(self):
        # start by setting up triggering on all compatible detectors
        self._set_det_trig(True)

        # As a convenience, make sure any LC400Buffer detectors are active
        from ..detectors.LC400Buffer import LC400Buffer
        inactive_buffs = []
        for d in LC400Buffer.getinstances():
            if not d.active:
                inactive_buffs.append(d)
        self._set_det_active(inactive_buffs, True)

        # configure the SC device - roughly like this
        axismap = {'sz': 1, 'sx': 2, 'sy': 3}
        fast_axis = axismap[self.fastmotor.name]
        import PyTango
        self.scancontrol = PyTango.DeviceProxy("B303A/CTL/FLYSCAN-02")
        ### note: these Tango attributes might have different names, check in Jive
        self.scancontrol.write_attribute("FlyScanMotorStartPosition", self.fastlimits[0])
        self.scancontrol.write_attribute("FlyScanMotorEndPosition", self.fastlimits[1])
        self.scancontrol.write_attribute("NumberOfIntervals", self.fastintervals)
        self.scancontrol.write_attribute("GateWidth", self.exptime)
        self.scancontrol.write_attribute("GateLatency", self.latency)
        self.scancontrol.write_attribute("FlyScanMotorAxis", self.fastmotor.axis)
        self.scancontrol.ConfigureLC400Motion()
        self.scancontrol.ConfigureLC400Recorder()
        self.scancontrol.ConfigureStanford()

        # run the main part
        super(NpointFlyscan, self).run()

        # deactivate position buffer detectors that were not active before
        self._set_det_active(inactive_buffs, False)

        # set back the triggering state
        self._set_det_trig(False)

    def _before_start(self):
        # we need to call Go() on the SC device at some point, maybe here.
        self.scancontrol.Go()

    def _while_acquiring(self):
        s = ''
        for d in Detector.get_active_detectors():
            if d.name in ('xspress3', 'pilatus', 'pilatus1m', 'merlin'):
                s += ('%s: %u, ' % (d.name, d.lima.last_image_acquired))
        print(s + '\r', end='')




from .Detector import Detector

import time
import numpy as np
import PyTango

class LC400Buffer(Detector):
    """
    Class representing the LC400 piezo machine under the
    control of the LC400ScanControl Tango device, used for
    reading the flyscan positions.
    """

    def __init__(self, name=None, device=None, xaxis=2, yaxis=3, zaxis=1):
        self.proxy = PyTango.DeviceProxy(device)
        Detector.__init__(self, name=name)
        self.xaxis = xaxis
        self.yaxis = yaxis
        self.zaxis = zaxis

    def initialize(self):
        self.proxy.init()

    def stop(self):
        self.proxy.Stop()

    def busy(self):
        return not (self.proxy.State() in (PyTango.DevState.STANDBY, PyTango.DevState.ON))

    def read(self):
        self.proxy.ReadLC400Buffer()
        data = {1: self.proxy.Axis1Positions,
                2: self.proxy.Axis2Positions,
                3: self.proxy.Axis3Positions,}
        return {'x': data[self.xaxis], 'y': data[self.yaxis], 'z': data[self.zaxis]}

    def start(self):
        """
        Placeholder, this detector just reads out whatever buffer is on the
        scancontrol device. That device is managed manually from macros.
        """
        pass




from .Lima import LimaDetector
from ..environment import env

import numpy as np
import PyTango
import os

class Merlin(LimaDetector):
    def __init__(self, *args, **kwargs):
        super(Merlin, self).__init__(*args, **kwargs)
        self._hdf_path_base = 'entry_%04d/measurement/Merlin/data'

    def _initialize_det(self):
        self.energy = 10.0
        self.det.write_attribute('gain', 'HGM')
        self.det.write_attribute('depth', 'BPP24')

    @property
    def energy(self):
        return self.det.operatingEnergy
        
    @energy.setter
    def energy(self, value):
        self.det.write_attribute('operatingEnergy', value)

    def prepare(self, *args, **kwargs):
        super(Merlin, self).prepare(*args, **kwargs)
        if self.hw_trig:
            self.det.write_attribute('triggerStartType', "RISING_EDGE_TTL")
        else:
            self.det.write_attribute('triggerStartType', "INTERNAL")




from .Detector import Detector, SoftwareLiveDetector, TriggeredDetector, Link
from ..environment import env

import time
import numpy as np
import PyTango
import os

class LimaDetector(Detector, SoftwareLiveDetector, TriggeredDetector):
    EXT_TRG_MODE = "EXTERNAL_TRIGGER_MULTI"

    def __init__(self, name=None, lima_device=None, det_device=None):
        self.lima_device_name = lima_device
        self.det_device_name = det_device
        Detector.__init__(self, name=name)
        SoftwareLiveDetector.__init__(self)
        TriggeredDetector.__init__(self)

    def _initialize_det(self):
        """
        Initialize detector-specific properties such as cutoff energies
        and bias voltages. Can also be used for overriding Lima settings
        that were initialized by the base class.
        """
        pass

    @property
    def energy(self):
        return -1

    @energy.setter
    def energy(self, val):
        pass

    def initialize(self):
        self.lima = PyTango.DeviceProxy(self.lima_device_name)
        self.lima.set_timeout_millis(10000)
        self.det = PyTango.DeviceProxy(self.det_device_name)

        # Make sure the devices are reachable, or this will throw an error
        self.lima.state()
        self.det.state()
        
        self.lima.acq_trigger_mode = "INTERNAL_TRIGGER"
        self.lima.saving_mode = "AUTO_FRAME"
        self.lima.saving_frame_per_file = 1
        self.lima.acq_nb_frames = 1
        self.lima.saving_managed_mode = 'SOFTWARE'
        self.lima.saving_overwrite_policy = 'MULTISET'
        self.lima.saving_format = 'HDF5'
        self.lima.saving_suffix = '.hdf5'
        self.lima.saving_index_format = ''
        self.lima.latency_time = 0.0
        self._initialize_det()

    def prepare(self, acqtime, dataid):
        """
        Run before acquisition, once per scan. Set up triggering,
        number of images etc.
        """
        # get out of the fault caused by trigger timeout
        if (self.lima.acq_status == 'Fault') and (self.lima.acq_status_fault_error == 'No error'):
            self.lima.stopAcq()
            self.lima.prepareAcq()

        if self.busy():
            raise Exception('%s is busy!' % self.name)

        if dataid is None:
            # no saving
            self.lima.saving_mode = "MANUAL" # no saving
            self.saving_filename = None
        else:
            # saving
            self.lima.saving_directory = env.paths.directory
            self.lima.saving_mode = "AUTO_FRAME"
            prefix = 'scan_%06d_%s' % (dataid, self.name)
            self.lima.saving_prefix = prefix
            self.saving_filename = prefix + self.lima.saving_suffix
            if os.path.exists(self.saving_filename):
                raise Exception('%s hdf5 file already exists' % self.name)

        if self.hw_trig:
            self.lima.acq_trigger_mode = self.EXT_TRG_MODE
            self.lima.acq_nb_frames = self.hw_trig_n
        else:
            self.lima.acq_trigger_mode = "INTERNAL_TRIGGER"
            self.lima.acq_nb_frames = 1

        self.image_number = -1
        self.acqtime = acqtime
        self.lima.acq_expo_time = acqtime

        while self.busy():
            print(self.name, 'slept 5 ms while waiting for prepare')
            time.sleep(.005)

    def arm(self):
        """
        Start the detector if hardware triggered, just prepareAcq otherwise.
        """
        if self.busy():
            raise Exception('%s is busy!' % self.name)
        self.image_number += 1
        self.lima.prepareAcq()
        while self.busy():
            time.sleep(.005)
        if self.hw_trig:
            self.lima.startAcq()

    def start(self):
        """
        Start acquisition when software triggered.
        """
        if self.hw_trig:
            return
        if self.busy():
            raise Exception('%s is busy!' % self.name)
        self.lima.startAcq()

    def stop(self):
        self.lima.stopAcq()
        self.stop_live()
        self.lima.stopAcq()

    def busy(self):
        return not self.lima.ready_for_next_acq

    def read(self):
        if self.saving_filename is None:
            return None
        else:
            absfile = os.path.join(self.lima.saving_directory, self.saving_filename)
            return Link(absfile , self._hdf_path_base % self.image_number)




"""
Provides a Motor subclass for the PI E727 piezo driver.
"""

import PyTango
from . import Motor

class E727Motor(Motor):
    """
    Single axis on the E727.
    """

    def __init__(self, device, axis, **kwargs):
        super(E727Motor, self).__init__(**kwargs)
        assert axis in (1, 2, 3)
        self.proxy = PyTango.DeviceProxy(device)
        self.axis = axis
        if axis == 1:
            self._mvrelfunc = self.proxy.move_relative1
        elif axis == 2:
            self._mvrelfunc = self.proxy.move_relative2
        elif axis == 3:
            self._mvrelfunc = self.proxy.move_relative3

    def _server_position(self):
        if self.axis == 1:
            return self.proxy.position1
        elif self.axis == 2:
            return self.proxy.position2
        elif self.axis == 3:
            return self.proxy.position3

    @property
    def dial_position(self):
        return = self._server_position()

    @dial_position.setter
    def dial_position(self, pos):
        current = self._server_position()
        self._mvrelfunc(pos - current)

    def busy(self):
        state = self.proxy.State()
        if state in (PyTango.DevState.STANDBY, PyTango.DevState.ON):
            return False
        return True

    def stop(self):
        self.proxy.stop()



"""
Provides a Motor subclass for Smaracts.
"""

import PyTango
from . import Motor

class SmaractLinearMotor(Motor):
    """
    Single Smaract motor axis.
    """

    def __init__(self, device, axis, **kwargs):
        super(SmaractLinearMotor, self).__init__(**kwargs)
        self.proxy = PyTango.DeviceProxy(device)
        self.axis = int(axis)

    @property
    def dial_position(self):
        return self.proxy.Read_position(self.axis)

    @dial_position.setter
    def dial_position(self, pos):
        self.proxy.Write_position('%d, %f' % (self.axis, pos))

    def busy(self):
        state = self.proxy.channel_status(self.axis)
        # Smaracts return this:
        # 0 (ON), 1 (MOVING), 2 (ALARM)
        return not (state == 0)

    def stop(self):
        self.proxy.stop_all() # safety first



try:
    import PyTango
except ModuleNotFoundError:
    pass

class PathFixer(object):
    """
    Basic pathfixer which takes a path manually.
    """
    def __init__(self):
        self.directory = None

class SdmPathFixer(object):
    """
    MAX IV pathfixer which takes a path from a Tango device.
    """
    def __init__(self, sdm_device):
        self.device = PyTango.DeviceProxy(sdm_device)

    @property
    def directory(self):
        return self.device.SamplePath




"""
This module contains the "environment" -- the set of objects that define how
and when data acquisition is done, what is seen, where data is written, etc.
It also provides and keeps track of macros.

The module provides a central instance of the Environment class, env.
"""

from IPython import get_ipython
ipython = get_ipython()

from .. import utils
from .data import PathFixer
from .scheduling import DummyScheduler
from .snapshots import MotorSnapshot

class Env(object):
    """
    Container for a number of global environment variables.
    """
    def __init__(self):
        self.registeredMacros = {}
        self.nextScanID = 0
        self.userLevel = 5
        self.paths = PathFixer()
        self.scheduler = DummyScheduler()
        self.snapshot = MotorSnapshot()
env = Env()

def runCommand(line):
    """
    Function for running magics from scripts.
    """
    ipython.magic(line)

def macro(cls):
    """
    Decorator which turn a class into a macro. This means that it will
    be callable from the ipython prompt with command line syntax.

    * Take all arguments in the constructor. The arguments can be Gadget
      objects and this decorator will find the objects from their names
      as entered on the command line. They can also be python expressions
      like 1+1 or 1./20.
    * Provide a run method which takes no arguments and executes the 
      scan or whatever.

    The name of the class converted to lower case will be used for the
    magic command used to launch the macro.
    """
    if not ipython: return cls

    def fcn(line):
        args = utils.str_to_args(line)
        try:
            obj = cls(*args)
        except MacroSyntaxError:
            print('Bad input. Usage:')
            print(cls.__doc__)
            return
        obj.run()
    if cls.__doc__:
        fcn.__doc__ = cls.__doc__
    else:
        print('Please document your macros! %s is missing a docstring.'\
                % cls.__name__)

    # sanity check
    assert cls.run.__call__

    name = cls.__name__.lower()
    env.registeredMacros[name] = cls
    ipython.register_magic_function(fcn, magic_name=name)
    return cls

class MacroSyntaxError(Exception):
    pass

@macro
class LsMac(object):
    """
    List available macros. Do <macro-name>? (without <>) for more information.
    """
    def run(self):
        print(utils.dict_to_table(env.registeredMacros, titles=('name', 'class'), sort=True))
        print('\nDo <macro-name>? (without <>) for more information.')

@macro
class UserLevel(object):
    """
    Get or set the current user level

    userlevel [<level>]
    """
    def __init__(self, level=None):
        if level is None:
            print("Current userlevel: %d" % env.userLevel)
        else:
            env.userLevel = level
    def run(self):
        pass





import sys
assert sys.version_info.major == 3, 'Use ipython3 instead.'

from . import recorders
from . import detectors
from . import motors
from . import scans
from . import Gadget
from . import utils
from . import environment



"""
Provides a Motor subclass for the Npoint LC400 piezo driver.
"""

import PyTango
from . import Motor

class LC400Motor(Motor):
    """
    Single axis on the LC400.
    """

    def __init__(self, device, axis, **kwargs):
        super(LC400Motor, self).__init__(**kwargs)
        assert axis in (1, 2, 3)
        self.proxy = PyTango.DeviceProxy(device)
        self.axis = axis
        self._format = '%.3f'

    @property
    def dial_position(self):
        if self.axis == 1:
            val = self.proxy.axis1_position
        elif self.axis == 2:
            val = self.proxy.axis2_position
        elif self.axis == 3:
            val = self.proxy.axis3_position
        return val

    @dial_position.setter
    def dial_position(self, pos):
        if self.axis == 1:
            self.proxy.axis1_position = pos
        elif self.axis == 2:
            self.proxy.axis2_position = pos
        elif self.axis == 3:
            self.proxy.axis3_position = pos

    def busy(self):
        attribute = 'axis%d_position_on_target' % self.axis
        on_target = self.proxy.read_attribute(attribute).value
        return not on_target

    def stop(self):
        self.proxy.stop_waveform()
        self.proxy.stop_recording()



from ..Gadget import Gadget
from ..environment import macro
from .. import utils
import numpy as np
import time
import threading
from h5py import ExternalLink

class Detector(Gadget):
    """
    Base class representing any device which can be read out to produce
    recordable data.
    """

    def __init__(self, *args, **kwargs):
        super(Detector, self).__init__(*args, **kwargs)
        try:
            self.active = True
            self.initialize()
        except Exception as e:
            self.active = False
            print('Failed to initialize %s. Run "%s.initialize()" to try again and perhaps learn more.' % (self.name, self.name))

    @classmethod
    def get_active_detectors(cls):
        """
        Returns a DetectorGroup instance containing the currently active
        detectors.
        """
        return DetectorGroup(*[d for d in cls.getinstances() if d.active])
    
    def prepare(self, acqtime, dataid):
        """
        Run before acquisition, once per scan.
            acqtime: exposure time for which to prepare
            dataid:  some way of identifying the data to
                     collected, useful for detectors that
                     write their own data files.
        """
        if self.busy():
            raise Exception('%s is busy!' % self.name)
        self.acqtime = acqtime

    def arm(self):
        """
        Run before every acquisition. Arm any hardware triggered detectors.
        """
        if self.busy():
            raise Exception('%s is busy!' % self.name)

    def start(self):
        """
        Start acquisition for any software triggered detectors.
        """
        if self.busy():
            raise Exception('%s is busy!' % self.name)

    def initialize(self):
        """
        Mandatory method for initializing a detector.
        """
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError

    def busy(self):
        raise NotImplementedError

    def read(self):
        raise NotImplementedError

class LiveDetector(object):
    """
    Abstract class to define the interface of live detectors, which
    can run continuously with no synchronization or data capture.
    """
    def __init__(self):
        pass

    def start_live(self, acqtime=1.0):
        raise NotImplementedError

    def stop_live(self):
        """
        This method should be made harmless, so that it can be
        run even if the detector is not running.
        """
        raise NotImplementedError

class SoftwareLiveDetector(LiveDetector):
    """
    Implements software live mode.
    """
    def __init__(self):
        self.thread = None
        self.stopped = False

    def start_live(self, acqtime=1.0):
        if self.thread is not None: self.stop_live()
        self.thread = threading.Thread(target=self._start, args=(acqtime,))
        self.thread.start()

    def stop_live(self):
        if self.thread is None: return
        self.stopped = True
        self.thread.join()
        self.thread = None

    def _start(self, acqtime):
        self.stopped = False
        self.prepare(acqtime, None)
        while not self.stopped:
            self.arm()
            self.start()
            while self.busy():
                time.sleep(.05)

class TriggeredDetector(object):
    """
    Defines the API for detectors that optionally accept hardware
    triggers.
    """
    def __init__(self):
        self.hw_trig = False
        self.hw_trig_n = 1

class DetectorGroup(object):
    """
    Collection of Detector objects to be acquired together, in a scan
    for example. Convenience class to call prepare, arm, busy etc
    in shorthand.
    """

    def __init__(self, *args):
        self.detectors = list()
        for arg in args:
            self.detectors.append(arg)

    def prepare(self, acqtime, dataid):
        for d in self:
            d.prepare(acqtime, dataid)

    def arm(self):
        for d in self:
            d.arm()

    def start(self):
        for d in self:
            d.start()

    def stop(self):
        for d in self:
            d.stop()

    def busy(self):
        for d in self:
            if d.busy(): return True
        return False

    def __iter__(self):
        return self.detectors.__iter__()

    def __len__(self):
        return self.detectors.__len__()

class Link(ExternalLink):
    """
    Some detectors write their own data to disk. When that happens,
    return data should be a link to where the actual data was saved.
    This class represents such links and is a simple wrapper around
    h5py.ExternalLink, with the added shape and type info which
    allows creating nice virtual datasets.
    """
    def __init__(self, filename, path, shape, dtype):
        self.shape = shape
        self.dtype = dtype
        super(Link, self).__init__(filename, path)

    def __str__(self):
        return '<link>'

@macro
class LsDet(object):
    """
    List available detectors.
    """
    def run(self):
        dct = {}
        for d in Detector.getinstances():
            name = ('* ' + d.name) if d.active else ('  ' + d.name)
            dct[name] = d.__class__
        print(utils.dict_to_table(dct, titles=('  name', 'class')))

@macro
class StartLive(object):
    """
    Starts software live mode on listed eligible detectors. If none
    are listed, all active and eligible detectors are started.

    startlive [<det1> ... <detN> <exposure time>]
    """
    def __init__(self, *args):
        try:
            self.exptime = float(args[-1])
            self.dets = args[:-1]
        except (TypeError, IndexError):
            self.exptime = .1
            self.dets = args
        if not self.dets:
            self.dets = [d for d in Detector.get_active_detectors() if isinstance(d, LiveDetector)]

    def run(self):
        for d in self.dets:
            if isinstance(d, LiveDetector):
                d.start_live(float(self.exptime))
            else:
                print('%s is not a LiveDetector' % d.name)

@macro
class StopLive(object):
    """
    Stops software live mode on listed eligible detectors. If
	no arguments are given, all active live detectors are
	stopped.

    stoplive [<det1> ... <detN>]
    """
    def __init__(self, *args):
        if args:
            self.dets = args
        else:
            self.dets = [d for d in Detector.get_active_detectors() if isinstance(d, LiveDetector)]

    def run(self):
        for d in self.dets:
            if isinstance(d, LiveDetector):
                d.stop_live()
            else:
                print('%s is not a LiveDetector' % d.name)

@macro
class Deactivate(object):
    """
    Deactivates all detectors or those specified.

    deactivate [<det1> ... <detN>]
    """
    def __init__(self, *args):
        if args:
            self.dets = args
        else:
            self.dets = Detector.getinstances()

    def run(self):
        for d in self.dets:
            d.active = False

@macro
class Activate(object):
    """
    Activates all detectors or those specified.

    deactivate [<det1> ... <detN>]
    """
    def __init__(self, *args):
        if args:
            self.dets = args
        else:
            self.dets = Detector.getinstances()

    def run(self):
        for d in self.dets:
            d.active = True




import weakref

class Gadget(object):
    """
    Base class for motors, detectors, etc. Main purpose is to keep track
    of instances so that names can be mapped to objects.

    This base class keeps a master class attribute holding references to
    all instances. The class method getinstances() on each subclass then
    filters and returns instances of that class or its children.

    So:
    [g.name for g in Gadget.getinstances()]         # is a list of all motors, detectors, and all else.
    [m.name for m in Motor.getinstances()]          # is a list of all motors.
    [m.name for m in DummyMotor.getinstances()]     # is a list of all dummy motors.

    """

    _base_class_instances = set()

    def __init__(self, name=None, userlevel=1):
        if not str(name) == name:
            raise Exception('Gadgets must have names!')
        self.name = name
        self.userlevel = userlevel
        self._base_class_instances.add(weakref.ref(self))

    @classmethod
    def getinstances(cls):
        """
        Smart method stolen from effbot.org and modified.
        """
        dead = set()
        for ref in cls._base_class_instances:
            obj = ref()
            if isinstance(obj, cls):
                yield obj
            elif obj is None:
                dead.add(ref)
        cls._base_class_instances -= dead



"""
Provides a Detector interface to Tango attributes, so that anything
can be monitored during scans.
"""

from .Detector import Detector
import PyTango

class TangoAttributeDetector(Detector):
    """
    Detector interface to Tango attributes, so that anything can be
    monitored during scans. These detectors simply take snapshots
    of Tango attributes and are never busy.
    """
    def __init__(self, name, device, attribute):
        super(TangoAttributeDetector, self).__init__(name=name)
        self.device_name = device
        self.attribute = attribute

    def initialize(self):
        self.proxy = PyTango.DeviceProxy(self.device_name)

    def start(self):
        super(TangoAttributeDetector, self).start()
        self.val = self.proxy.read_attribute(self.attribute).value

    def stop(self):
        pass

    def busy(self):
        return False

    def read(self):
        try:
            return self.val
        except AttributeError:
            raise Exception('Detector not started!')




from . import Recorder
from ..environment import macro

import signal
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D


class PlotRecorder(Recorder):
    """
    Recorder which catches data and plots it with matplotlib.

    Unlike the base class Recorder, the GUI event loop takes care of
    the timing (plt.timer) and when to close (plt.show).
    """

    def __init__(self, data1, data2=None, name='plot'):
        Recorder.__init__(self, name=name)
        if data2 is not None:
            self.xdata = data1
            self.ydata = data2
        else:
            self.xdata = None
            self.ydata = data1
        self.nplots = 0

    def init(self):
        # set up figure and axes
        self.fig = plt.figure()
        self.fig.canvas.set_window_title(self.name)
        self.ax = self.fig.gca()
        self.ax.set_xlabel(self.xdata)
        self.ax.set_ylabel(self.ydata)

        # add a timer to trigger periodic checking of the queue
        self.timer = self.fig.canvas.new_timer(interval=int(self.delay*1000))
        self.timer.add_callback(self._timer_callback)
        self.timer.start()

        # blocking show() manages when the application should close
        plt.show()

    def _timer_callback(self):
        # this SIGINT handling has to be set up here, for some reason
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        self._process_queue()
        self.periodic_check()

    def run(self):
        self.init()

    def act_on_header(self, dct):
        # start a new scan
        self.nplots += 1
        self.x = []
        self.new_scan = True
        self.scannr = dct['scannr']

    def act_on_data(self, dct):
        # if our data isn't in dct, just move on
        try:
            new_data = dct[self.ydata]
        except KeyError:
            return

        # be ready to get data that are dicts instead of numbers,
        # so may as well just work with dicts.
        if not isinstance(new_data, dict):
            new_data = {'': new_data}

        # we can only set up lines etc once we know what is actually
        # in the data.
        if self.new_scan:
            self.new_scan = False
            col = 'bkmrcg'[(self.nplots-1) % 6]
            styles = {k:['solid', 'dashed', 'dotted', 'dashdot'][i%4] for i, k in enumerate(new_data.keys())}
            self.lines = {key: Line2D(xdata=[], ydata=[], color=col, linestyle=styles[key], label='%d: %s'%(self.scannr, key)) for key in new_data.keys()}
            self.y = {key: [] for key in new_data.keys()}
            for k, l in self.lines.items():
                self.ax.add_line(l)
            self.ax.legend()

        # ok treat the actual data
        for k, v in new_data.items():
            self.y[k].append(v)
        if self.xdata is not None:
            # checking if we have explicit x data
            self.x.append(dct[self.xdata])
        else:
            self.x = range(len(self.y[k]))
        for k, l in self.lines.items():
            l.set_data(self.x, self.y[k])

        self.ax.relim()
        self.ax.autoscale_view()
        plt.draw()

    def _close(self):
        plt.close(self.fig)

@macro
class LivePlot(object):
    """
    Start a live plot recorder which will plot coming scans.

    liveplot [<x>] <y>

    Examples:
    liveplot xmotor diode1
    liveplot diode1
    """
    def __init__(self, data1, data2=None):
        basename = 'plot'
        name = basename
        i = 2
        while name in [r.name for r in Recorder.getinstances()]:
            name = basename + '_%d' % i
            i += 1
        self.data1 = data1.name if hasattr(data1, 'name') else data1
        self.data2 = data2.name if hasattr(data2, 'name') else data2
        self.name = name
    def run(self):
        rec = PlotRecorder(data1=self.data1, data2=self.data2, name=self.name)
        rec.start()



from .Motor import Motor, DummyMotor, MotorMemorizer
from .PseudoMotor import PseudoMotor, ExamplePseudoMotor

def all_are_motors(seq):
    """
    Function which returns True if all objects in seq are instances
    of Motor or its subclasses.
    """
    checks = [isinstance(m, Motor) for m in seq]
    return not (False in checks)

# don't automatically import hardware-specific classes here,
# as these might have special dependencies that aren't
# available everywhere.



from .Detector import Detector, LiveDetector, TriggeredDetector
import PyTango
import time

PROP_MAP = {1: 'counter01', 2: 'counter02', 3: 'counter03'}
CHANNEL_MAP = {'counter1': 'CNT1', 'counter2': 'CNT2', 'counter3': 'CNT3'}

class Ni6602CounterCard(Detector, LiveDetector, TriggeredDetector):
    """
    Interface to the Ni6602 Tango device exposing counters.
    """
    def __init__(self, name=None, device=None):
        self.dev_name = device
        Detector.__init__(self, name=name)
        LiveDetector.__init__(self)
        TriggeredDetector.__init__(self)

    def initialize(self):
        self.dev = PyTango.DeviceProxy(self.dev_name)
        self.dev.init()
        
        # handle some unwanted options that otherwise can cause errors
        self.dev.write_attribute("nexusFileGeneration", 0)
        self.dev.write_attribute("CNT1MinPulseWidthEnable", 0)
        self.dev.write_attribute("CNT2MinPulseWidthEnable", 0)
        self.dev.write_attribute("CNT3MinPulseWidthEnable", 0)

    def _toggle_PW_mode(self, pw):
        """
        Switch between PW (pulse-width gated) and EVT (single count) mode.
        This is done through Tango properties so is a bit messy.

        pw = True means go to gated mode, 
        pw = False means go to single count.
        """
        if pw:
            oldstr, newstr = 'Mode:EVT', 'Mode:PW'
        else:
            oldstr, newstr = 'Mode:PW', 'Mode:EVT'
        for prop_name in PROP_MAP.values():
            channel_props = self.dev.get_property(prop_name)[prop_name]
            for i, val in enumerate(channel_props):
                if val == oldstr:
                    channel_props[i] = newstr
                    self.dev.put_property({prop_name: channel_props})
                    break
            time.sleep(.1)
            self.dev.init()

    def prepare(self, acqtime, dataid):
        """
        Run before acquisition, once per scan. Set up triggering,
        number of images etc.
        """

        if self.busy():
            raise Exception('%s is busy!' % self.name)

        if self.hw_trig:
            self._toggle_PW_mode(True)
            self.dev.acquisitionMode = 'BUFFERED'
            self.dev.totalNbPoint = self.hw_trig_n
            self.dev.bufferDepth = self.hw_trig_n
        else:
            self._toggle_PW_mode(False)
            self.dev.acquisitionMode = 'SCALAR'
        self.dev.integrationTime = acqtime

    def start_live(self, acqtime=1.0):
        """
        The NI card can do on-board live mode.
        """
        self.dev.continuous = True
        self.dev.integrationTime = acqtime
        self.dev.Start()

    def stop_live(self):
        """
        The NI card can do on-board live mode.
        """
        self.stop()
        self.dev.continuous = False

    def arm(self):
        """
        Start the detector if hardware triggered
        """
        if self.busy():
            raise Exception('%s is busy!' % self.name)
        if self.hw_trig:
            self.dev.Start()

    def start(self):
        """
        Start acquisition when software triggered.
        """
        if self.hw_trig:
            return
        if self.busy():
            raise Exception('%s is busy!' % self.name)
        self.dev.Start()

    def stop(self):
        self.dev.Stop()
        while not self.dev.State() == PyTango.DevState.STANDBY:
            time.sleep(.01)

    def busy(self):
        return not (self.dev.State() == PyTango.DevState.STANDBY)

    def read(self):
        res = {}
        for name, channel in CHANNEL_MAP.items():
            res[name] = self.dev.read_attribute(channel).value
        return res




"""
Provides a Motor interface to Tango motors.
"""

import PyTango
from . import Motor

class TangoMotor(Motor):
    """
    Single motor as exposed by Pool. Use for IcePAP:s.
    """

    def __init__(self, device, **kwargs):
        super(TangoMotor, self).__init__(**kwargs)
        self.proxy = PyTango.DeviceProxy(device)

    @property
    def dial_position(self):
        return self.proxy.Position

    @dial_position.setter
    def dial_position(self, pos):
        self.proxy.Position = pos

    def busy(self):
        state = self.proxy.State()
        return not (state == PyTango.DevState.ON)

    def stop(self):
        self.proxy.stop()

